# ruff: noqa: SLF001
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock

from gendisc.utils import (
    DirectorySplitter,
    LazyMounts,
    WriteSpeeds,
    get_dir_size,
    get_disc_type,
    is_cross_fs,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


@pytest.fixture
def mocker_fs(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.walk', return_value=[('basepath', [], ['file1', 'file2'])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_file_size', return_value=1024)
    mocker.patch('gendisc.utils.Path')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='dir1\ndir2\n'))


def test_get_disc_type() -> None:
    assert get_disc_type(700 * 1024 * 1024) == 'DVD-R'
    assert get_disc_type(int(4.7 * 1024 * 1024 * 1024)) == 'DVD-R DL'
    assert get_disc_type(int(8.5 * 1024 * 1024 * 1024)) == 'BD-R'
    assert get_disc_type(25 * 1024 * 1024 * 1024) == 'BD-R DL'
    assert get_disc_type(50 * 1024 * 1024 * 1024) == 'BD-R XL (100 GB)'
    assert get_disc_type(100 * 1024 * 1024 * 1024) == 'BD-R XL (128 GB)'
    with pytest.raises(ValueError, match=r'Disc size exceeds maximum supported size.'):
        get_disc_type(128 * 1024 * 1024 * 1024)


def test_get_dir_size() -> None:
    with pytest.raises(NotADirectoryError):
        get_dir_size('non-existent-path')


def test_get_dir_size_returns_correct_size(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], ['a', 'b', 'c'])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_file_size', return_value=2048)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    size = get_dir_size('some_dir')
    assert size == 3 * 2048


def test_get_dir_size_skips_symlinks(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], ['a', 'b'])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', side_effect=[False, True])
    mocker.patch('gendisc.utils.get_file_size', return_value=4096)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    size = get_dir_size('dir')
    assert size == 4096


def test_get_dir_size_handles_oserror(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], ['z', 'x'])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_file_size', side_effect=[OSError, 512, 512, 512])
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    size = get_dir_size('dir2')
    assert size == 1536


def test_get_dir_size_raises_not_a_directory(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.isdir', return_value=False)
    with pytest.raises(NotADirectoryError):
        get_dir_size('not_a_dir')


def test_get_dir_size_reports_buggy_fs_once(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils._REPORTED_BUGGY_FS', False)  # noqa: FBT003
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mock_log_warning = mocker.patch('gendisc.utils.log.warning')
    mocker.patch('gendisc.utils.walk',
                 side_effect=[[('base', [], ['a'])], [('base', [], ['a'])], [('base', [], ['a'])],
                              [('base', [], ['a'])]])
    mocker.patch('gendisc.utils.get_file_size', side_effect=[OSError, 2048, OSError, 2048])
    get_dir_size('dir')
    mock_log_warning.assert_called_once_with(
        'Buggy file system (cifs with "unix" option?) reported directory %s as file.', 'base/a')
    mock_log_warning.reset_mock()
    get_dir_size('dir')
    assert mock_log_warning.call_count == 0


def test_get_dir_size_reports_oserror_exception(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.isdir', side_effect=[True, False])
    mock_log_exception = mocker.patch('gendisc.utils.log.exception')
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], ['a'])])
    mocker.patch('gendisc.utils.get_file_size', side_effect=OSError)
    get_dir_size('dir')
    mock_log_exception.assert_called_once_with(
        'Caught error getting file size for %s. It will not be considered part of the total.',
        'base/a')


def test_write_speeds_defaults() -> None:
    speeds = WriteSpeeds()
    assert speeds.cd == 24
    assert speeds.dvd == 8
    assert speeds.dvd_dl == 8
    assert speeds.bd == 4
    assert speeds.bd_dl == 6
    assert speeds.bd_tl == 4
    assert speeds.bd_xl == 4


def test_write_speeds_get_speed_valid() -> None:
    speeds = WriteSpeeds()
    assert speeds.get_speed('CD-R') == 24
    assert speeds.get_speed('DVD-R') == 8
    assert speeds.get_speed('DVD-R DL') == 8
    assert speeds.get_speed('BD-R') == 4
    assert speeds.get_speed('BD-R DL') == 6
    assert speeds.get_speed('BD-R XL (100 GB)') == 4
    assert speeds.get_speed('BD-R XL (128 GB)') == 4


def test_write_speeds_get_speed_invalid() -> None:
    speeds = WriteSpeeds()
    with pytest.raises(ValueError, match=r'Unknown disc type:'):
        speeds.get_speed('UNKNOWN-DISC')  # type: ignore[arg-type]


def test_is_cross_fs(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.utils.MOUNTS', ['/', '/mnt'])
    assert is_cross_fs('/') is True
    assert is_cross_fs('/mnt') is True
    assert is_cross_fs('/home') is False


def test_directory_splitter_init(mocker_fs: None) -> None:
    splitter = DirectorySplitter('test_path', 'prefix')
    assert splitter._prefix == 'prefix'
    assert splitter._delete_command == 'trash'
    assert splitter._drive == '/dev/sr0'
    assert splitter._starting_index == 1


def test_directory_splitter_split(mocker: MockerFixture, mocker_fs: None) -> None:
    mock_path = mocker.patch('gendisc.utils.Path')
    mock_path.reset_mock()
    mock_write_text = (
        mock_path.return_value.__truediv__.return_value.__truediv__.return_value.write_text)
    splitter = DirectorySplitter('test_path',
                                 'prefix-' * 10,
                                 preparer='preparer',
                                 publisher='publisher')
    splitter.split()
    assert len(splitter._sets) == 1
    assert len(splitter._sets[0]) == 1
    # Test the prefix is truncated
    shell = mock_write_text.call_args_list[1].args[0]
    assert 'VOLID=prefix-prefix-prefix-prefix-p-01' in shell
    assert '-preparer preparer -publisher publisher' in shell


def test_directory_splitter_too_large(mocker_fs: None) -> None:
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter._size = 1024
    splitter._too_large()
    assert splitter._total == 0
    assert len(splitter._current_set) == 0


def test_directory_splitter_append_set(mocker_fs: None) -> None:
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter._current_set = ['file1', 'file2']
    splitter._total = 2048
    splitter._append_set()
    assert len(splitter._sets) == 1
    assert len(splitter._sets[0]) == 2


def test_directory_splitter_split_skips_cross_fs(mocker: MockerFixture, mocker_fs: None) -> None:
    mock_write_spiral = mocker.patch('gendisc.utils.write_spiral_text_png')
    mocker.patch('gendisc.utils.shutil.which', return_value='fake-mogrify')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='.\ndir1\ndir2\n'))
    mock_path = mocker.patch('gendisc.utils.Path')
    mock_path.resolve.return_value = MagicMock(strict=True, parent=MagicMock())
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], ['file1'])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_file_size', return_value=1024)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    mocker.patch('gendisc.utils.is_cross_fs', side_effect=lambda d: d == 'dir2')
    splitter = DirectorySplitter('test_path', 'prefix', labels=True)
    splitter.split()
    assert len(splitter._sets) == 1
    assert all('dir1' in entry for entry in splitter._sets[0])
    assert all('dir2' not in entry for entry in splitter._sets[0])
    mock_write_spiral.assert_called_once_with(
        mock_path.return_value.__truediv__.return_value.__truediv__.return_value, 'prefix-01 || ')


def test_directory_splitter_split_file_too_large_for_bluray(mocker: MockerFixture,
                                                            mocker_fs: None) -> None:
    mocker.patch('gendisc.utils.shutil.which')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='.\nfile1\n'))
    mocker.patch('gendisc.utils.Path.resolve',
                 return_value=MagicMock(strict=True, parent=MagicMock()))
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], [])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_dir_size', side_effect=NotADirectoryError)
    mocker.patch('gendisc.utils.get_file_size', return_value=200 * 1024 * 1024 * 1024)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter.split()
    assert len(splitter._sets) == 0


def test_directory_splitter_split_file_too_large_for_bluray_already_xl(
        mocker: MockerFixture, mocker_fs: None) -> None:
    mocker.patch('gendisc.utils.shutil.which')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='.\nfile1\nfile2\n'))
    mocker.patch('gendisc.utils.Path.resolve',
                 return_value=MagicMock(strict=True, parent=MagicMock()))
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], [])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_dir_size',
                 side_effect=[101 * 1024 * 1024 * 1024, 101 * 1024 * 1024 * 1024])
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter.split()
    assert len(splitter._sets) == 2


def test_directory_splitter_split_file_too_large_for_bluray_tl_but_not_xl(
        mocker: MockerFixture, mocker_fs: None) -> None:
    mocker.patch('gendisc.utils.shutil.which')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='.\nfile1\n'))
    mocker.patch('gendisc.utils.Path.resolve',
                 return_value=MagicMock(strict=True, parent=MagicMock()))
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], [])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_dir_size', side_effect=NotADirectoryError)
    mocker.patch('gendisc.utils.get_file_size', return_value=100 * 1024 * 1024 * 1024)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter.split()
    assert len(splitter._sets) == 1


def test_directory_splitter_split_file_too_large_for_split_dir_separately(
        mocker: MockerFixture, mocker_fs: None) -> None:
    mock_log_debug = mocker.patch('gendisc.utils.log.debug')
    mocker.patch('gendisc.utils.shutil.which')
    mocker.patch('gendisc.utils.sp.run',
                 side_effect=[MagicMock(stdout='.\nfile1\n'),
                              MagicMock(stdout='.\n')])
    mocker.patch('gendisc.utils.Path.resolve',
                 return_value=MagicMock(strict=True, parent=MagicMock()))
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], [])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_dir_size', return_value=122 * 1024 * 1024 * 1024)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    splitter = DirectorySplitter('test_path', 'prefix')
    splitter.split()
    assert len(splitter._sets) == 0
    mock_log_debug.assert_has_calls(
        [mocker.call('Directory %s too large for Blu-ray. Splitting separately.', 'file1')])


def test_directory_splitter_skip_files_that_raise_oserror(mocker: MockerFixture,
                                                          mocker_fs: None) -> None:
    mocker.patch('gendisc.utils.shutil.which', return_value='fake-mogrify')
    mocker.patch('gendisc.utils.sp.run', return_value=MagicMock(stdout='.\ndir_big\n'))
    mocker.patch('gendisc.utils.Path.resolve',
                 return_value=MagicMock(strict=True, parent=MagicMock()))
    mocker.patch('gendisc.utils.walk', return_value=[('base', [], [])])
    mocker.patch('gendisc.utils.isdir', return_value=True)
    mocker.patch('gendisc.utils.islink', return_value=False)
    mocker.patch('gendisc.utils.get_dir_size', side_effect=NotADirectoryError)
    mocker.patch('gendisc.utils.get_file_size', side_effect=OSError)
    mocker.patch('gendisc.utils.path_join', side_effect=lambda base, f: f'{base}/{f}')
    recursive_called = {}
    orig_split = DirectorySplitter.split

    def fake_split(self: Any) -> None:
        recursive_called['called'] = True

    mocker.patch.object(DirectorySplitter, 'split', fake_split)
    splitter = DirectorySplitter('test_path', 'prefix', labels=True)
    orig_split(splitter)
    assert splitter._total == 0
    assert splitter._current_set == []


def test_lazy_mounts_read(mocker: MockerFixture) -> None:
    mock_mounts_content = '/dev/sda1 / ext4 rw 0 0\n/dev/sdb1 /mnt ext4 rw 0 0'
    mocker.patch('gendisc.utils.Path.read_text', return_value=mock_mounts_content)
    mounts = LazyMounts._read()
    assert mounts == ['/', '/mnt']


def test_lazy_mounts_initialize_and_reload(mocker: MockerFixture) -> None:
    mocker.patch.object(LazyMounts, '_read', return_value=['/mnt', '/media'])
    lm = LazyMounts()
    assert lm._mounts is None
    lm.initialize()
    assert lm._mounts == ['/mnt', '/media']
    lm._mounts = None
    lm.reload()
    assert lm._mounts == ['/mnt', '/media']


def test_lazy_mounts_mounts_property(mocker: MockerFixture) -> None:
    mocker.patch.object(LazyMounts, '_read', return_value=['/foo', '/bar'])
    lm = LazyMounts()
    mounts = lm.mounts
    assert mounts == ['/foo', '/bar']


def test_lazy_mounts_getitem_and_len(mocker: MockerFixture) -> None:
    mocker.patch.object(LazyMounts, '_read', return_value=['/a', '/b', '/c'])
    lm = LazyMounts()
    # __getitem__ int
    assert lm[0] == '/a'
    # __getitem__ slice
    assert lm[1:] == ['/b', '/c']
    # __len__
    assert len(lm) == 3



"""Generate disk file paths for mkisofs."""
from __future__ import annotations

from .genlabel import MogrifyNotFound, Point, write_spiral_text_png, write_spiral_text_svg
from .utils import DirectorySplitter

__all__ = ('DirectorySplitter', 'MogrifyNotFound', 'Point', 'write_spiral_text_png',
           'write_spiral_text_svg')
__version__ = '0.0.11'



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from gendisc.genlabel import (
    MogrifyNotFound,
    Point,
    _line_intersection,  # noqa: PLC2701
    create_spiral_path,
    create_spiral_text_svg,
    write_spiral_text_png,
    write_spiral_text_svg,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_write_spiral_text_png_success(mocker: MockerFixture) -> None:
    mock_write_svg = mocker.patch('gendisc.genlabel.write_spiral_text_svg')
    mock_run = mocker.patch('gendisc.genlabel.sp.run')
    mocker.patch('gendisc.genlabel.shutil.which', return_value='/usr/bin/mogrify')
    mock_exists = mocker.patch('gendisc.genlabel.Path.exists', return_value=True)
    mock_unlink = mocker.patch('gendisc.genlabel.Path.unlink')
    filename = 'test.png'
    text = 'spiral text'
    # Should not raise
    write_spiral_text_png(filename, text)
    mock_write_svg.assert_called_once()
    mock_run.assert_called_once()
    mock_exists.assert_called_once()
    mock_unlink.assert_called_once()


def test_write_spiral_text_png_keep_svg(mocker: MockerFixture) -> None:
    mock_write_svg = mocker.patch('gendisc.genlabel.write_spiral_text_svg')
    mocker.patch('gendisc.genlabel.sp.run')
    mocker.patch('gendisc.genlabel.shutil.which', return_value='/usr/bin/mogrify')
    mocker.patch('gendisc.genlabel.Path.exists', return_value=True)
    mock_unlink = mocker.patch('gendisc.genlabel.Path.unlink')
    write_spiral_text_png('file.png', 'txt', keep=True)
    mock_write_svg.assert_called_once()
    mock_unlink.assert_not_called()


def test_write_spiral_text_png_mogrify_not_found(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.genlabel.shutil.which', return_value=None)
    with pytest.raises(MogrifyNotFound):
        write_spiral_text_png('file.png', 'txt')


def test_write_spiral_text_png_file_not_created(mocker: MockerFixture) -> None:
    mocker.patch('gendisc.genlabel.write_spiral_text_svg')
    mocker.patch('gendisc.genlabel.sp.run')
    mocker.patch('gendisc.genlabel.shutil.which', return_value='/usr/bin/mogrify')
    mocker.patch('gendisc.genlabel.Path.exists', return_value=False)
    mocker.patch('gendisc.genlabel.Path.unlink')
    with pytest.raises(FileNotFoundError):
        write_spiral_text_png('file.png', 'txt')


def test_write_spiral_text_svg_writes_file(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('gendisc.genlabel.Path')
    mock_write_text = mock_path.return_value.write_text
    mock_create_svg = mocker.patch('gendisc.genlabel.create_spiral_text_svg',
                                   return_value='<svg>...</svg>')
    filename = 'spiral.svg'
    text = 'test spiral'
    # Should not raise
    write_spiral_text_svg(filename, text, width=123, height=456, font_size=22)
    mock_create_svg.assert_called_once_with(text, 123, 456, None, 22, None, 0, 40, -6840, 0, 30)
    mock_write_text.assert_called_once()
    args, kwargs = mock_write_text.call_args
    assert args[0].startswith('<svg')
    assert args[0].endswith('\n')
    assert kwargs['encoding'] == 'utf-8'


def test_write_spiral_text_svg_with_all_args(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('gendisc.genlabel.Path')
    mock_write_text = mock_path.return_value.write_text
    mock_create_svg = mocker.patch('gendisc.genlabel.create_spiral_text_svg',
                                   return_value='<svg>spiral</svg>')
    filename = 'spiral.svg'
    text = 'spiral'
    width = 200
    height = 300
    view_box = (0, 0, 400, 400)
    font_size = 18
    center = Point(1, 2)
    start_radius = 5
    space_per_loop = 10
    start_theta = -100
    end_theta = 100
    theta_step = 10
    write_spiral_text_svg(filename, text, width, height, view_box, font_size, center, start_radius,
                          space_per_loop, start_theta, end_theta, theta_step)
    mock_create_svg.assert_called_once_with(text, width, height, view_box, font_size, center,
                                            start_radius, space_per_loop, start_theta, end_theta,
                                            theta_step)
    mock_write_text.assert_called_once()
    args, kwargs = mock_write_text.call_args
    assert args[0].startswith('<svg')
    assert args[0].endswith('\n')
    assert kwargs['encoding'] == 'utf-8'


def test_write_spiral_text_svg_path_conversion(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('gendisc.genlabel.Path')
    mock_write_text = mock_path.return_value.write_text
    mocker.patch('gendisc.genlabel.create_spiral_text_svg', return_value='<svg>spiral</svg>')
    filename = 'file.svg'
    text = 'abc'
    write_spiral_text_svg(filename, text)
    mock_path.assert_any_call(filename)
    mock_write_text.assert_called_once()


def test_create_spiral_text_svg_defaults(mocker: MockerFixture) -> None:
    # Patch create_spiral_path to return a known path string
    mock_path = mocker.patch('gendisc.genlabel.create_spiral_path', return_value='M 0,0 Q 1,1 2,2')
    text = 'hello spiral'
    svg = create_spiral_text_svg(text)
    assert svg.startswith('<?xml')
    assert '<svg' in svg
    assert '<textPath' in svg
    assert text in svg
    mock_path.assert_called_once()


def test_create_spiral_text_svg_custom_args(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('gendisc.genlabel.create_spiral_path', return_value='M 1,2 Q 3,4 5,6')
    text = 'custom'
    width = 123
    height = 456
    view_box = (0, 0, 10, 20)
    font_size = 33
    center = Point(7, 8)
    start_radius = 9
    space_per_loop = 10
    start_theta = -100
    end_theta = 100
    theta_step = 15
    svg = create_spiral_text_svg(
        text,
        width=width,
        height=height,
        view_box=view_box,
        font_size=font_size,
        center=center,
        start_radius=start_radius,
        space_per_loop=space_per_loop,
        start_theta=start_theta,
        end_theta=end_theta,
        theta_step=theta_step,
    )
    assert f'width="{width}"' in svg
    assert f'height="{height}"' in svg
    assert f'font: {font_size}px' in svg
    assert 'viewBox="0 0 10 20"' in svg
    assert text in svg
    mock_path.assert_called_once_with(center, start_radius, space_per_loop, start_theta, end_theta,
                                      theta_step)


def test_create_spiral_text_svg_view_box_none(mocker: MockerFixture) -> None:
    # Should use default viewBox if not provided
    mocker.patch('gendisc.genlabel.create_spiral_path', return_value='M 0,0 Q 1,1 2,2')
    svg = create_spiral_text_svg('spiral', width=50, height=60)
    assert 'viewBox="0 0 100 120"' in svg


def test_create_spiral_text_svg_center_none(mocker: MockerFixture) -> None:
    # Should use (width, width) as center if not provided
    called_args = {}

    def fake_create_spiral_path(center: Point, *args: Any, **kwargs: Any) -> str:
        called_args['center'] = center
        return 'M 0,0'

    mocker.patch('gendisc.genlabel.create_spiral_path', side_effect=fake_create_spiral_path)
    width = 77
    create_spiral_text_svg('spiral', width=width)
    assert called_args['center'] == Point(width, width)


def test_create_spiral_path_defaults(mocker: MockerFixture) -> None:
    # Patch math.radians to identity for easier calculation
    mocker.patch('math.radians', side_effect=float)
    mocker.patch('math.cos', side_effect=lambda _: 1)
    mocker.patch('math.sin', side_effect=lambda _: 0)
    # Patch _line_intersection to return a fixed point
    mocker.patch('gendisc.genlabel._line_intersection', return_value=Point(1, 2))
    # Patch _p_str to just return the coordinates as string
    mocker.patch('gendisc.genlabel._p_str', side_effect=lambda p: f'{p.x},{p.y} ')
    path = create_spiral_path()
    assert path.startswith('M ')
    assert 'Q' in path
    assert isinstance(path, str)


def test_create_spiral_path_custom_args(mocker: MockerFixture) -> None:
    mocker.patch('math.radians', side_effect=float)
    mocker.patch('math.cos', side_effect=lambda _: 2)
    mocker.patch('math.sin', side_effect=lambda _: 3)
    mocker.patch('gendisc.genlabel._line_intersection', return_value=Point(5, 6))
    mocker.patch('gendisc.genlabel._p_str', side_effect=lambda p: f'{p.x}:{p.y} ')
    center = Point(10, 20)
    path = create_spiral_path(
        center=center,
        start_radius=2,
        space_per_loop=8,
        start_theta=-100,
        end_theta=100,
        theta_step=10,
    )
    assert path.startswith('M ')
    assert 'Q' in path
    assert isinstance(path, str)


def test_create_spiral_path_looping_and_path_content(mocker: MockerFixture) -> None:
    # Use real math for radians, cos, sin, but patch _line_intersection and _p_str
    path = create_spiral_path(
        center=Point(0, 0),
        start_radius=1,
        space_per_loop=2,
        start_theta=-60,
        end_theta=60,
        theta_step=30,
    )
    assert path.startswith('M ')
    assert 'Q' in path
    assert '1.0,0.0' in path


def test_create_spiral_path_parallel_lines_raises(mocker: MockerFixture) -> None:
    # Patch _line_intersection to raise ValueError for parallel lines
    def fake_line_intersection(m1: float, b1: float, m2: float, b2: float) -> Point:
        if m1 == m2:
            msg = 'Lines are parallel and do not intersect.'
            raise ValueError(msg)
        return Point(0, 0)

    mocker.patch('gendisc.genlabel._line_intersection', side_effect=fake_line_intersection)
    mocker.patch('math.radians', side_effect=float)
    mocker.patch('math.cos', side_effect=lambda _: 1)
    mocker.patch('math.sin', side_effect=lambda _: 0)
    mocker.patch('gendisc.genlabel._p_str', side_effect=lambda p: f'{p.x},{p.y} ')
    with pytest.raises(ValueError, match='parallel'):
        _line_intersection(1, 2, 1, 3)



"""Utilities."""
from __future__ import annotations

from collections.abc import Sequence
from functools import cache
from os import walk
from os.path import commonprefix, isdir, islink
from pathlib import Path
from typing import Literal, NamedTuple, overload
import logging
import logging.config
import os
import re
import shlex
import shutil
import subprocess as sp

from fsutil import get_file_size
from tqdm import tqdm
from typing_extensions import override
import fsutil
import jinja2

from .constants import (
    BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED,
    BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED,
    BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED,
    BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED,
    CD_R_BYTES_ADJUSTED,
    DVD_R_DUAL_LAYER_SIZE_BYTES_ADJUSTED,
    DVD_R_SINGLE_LAYER_SIZE_BYTES,
)
from .genlabel import write_spiral_text_png

__all__ = ('DirectorySplitter', 'WriteSpeeds', 'get_disc_type')

log = logging.getLogger(__name__)
_jinja_env = jinja2.Environment(autoescape=jinja2.select_autoescape(),
                                loader=jinja2.PackageLoader(__package__),
                                lstrip_blocks=True,
                                trim_blocks=True,
                                undefined=jinja2.StrictUndefined)


def setup_logging(*,
                  debug: bool = False,
                  force_color: bool = False,
                  no_color: bool = False) -> None:  # pragma: no cover
    """Set up logging configuration."""
    logging.config.dictConfig({
        'disable_existing_loggers': True,
        'root': {
            'level': 'DEBUG' if debug else 'INFO',
            'handlers': ['console'],
        },
        'formatters': {
            'default': {
                '()': 'colorlog.ColoredFormatter',
                'force_color': force_color,
                'format': (
                    '%(light_cyan)s%(asctime)s%(reset)s | %(log_color)s%(levelname)-8s%(reset)s | '
                    '%(light_green)s%(name)s%(reset)s:%(light_red)s%(funcName)s%(reset)s:'
                    '%(blue)s%(lineno)d%(reset)s - %(message)s'),
                'no_color': no_color,
            },
            'simple': {
                'format': '%(levelname)s: %(message)s',
            }
        },
        'handlers': {
            'console': {
                'class': 'colorlog.StreamHandler',
                'formatter': 'default' if debug else 'simple',
            }
        },
        'loggers': {
            'gendisc': {
                'level': 'INFO' if not debug else 'DEBUG',
                'handlers': ('console',),
                'propagate': False,
            },
            'wakepy': {
                'level': 'INFO' if not debug else 'DEBUG',
                'handlers': ('console',),
                'propagate': False,
            },
        },
        'version': 1
    })


convert_size_bytes_to_string = cache(fsutil.convert_size_bytes_to_string)
path_join = cache(os.path.join)
quote = cache(shlex.quote)

_REPORTED_BUGGY_FS = False


def get_dir_size(path: str) -> int:
    global _REPORTED_BUGGY_FS  # noqa: PLW0603
    size = 0
    if not isdir(path):  # noqa: PTH112
        raise NotADirectoryError
    for basepath, _, filenames in tqdm(walk(path), desc=f'Calculating size of {path}', unit=' dir'):
        for filename in filenames:
            filepath = path_join(basepath, filename)
            if not islink(filepath):  # noqa: PTH114
                try:
                    log.debug('Getting file size for %s.', filepath)
                    size += get_file_size(filepath)
                except OSError:
                    if isdir(filepath):  # noqa: PTH112
                        # On cifs with 'unix' option directories get reported as files from walk().
                        if not _REPORTED_BUGGY_FS:
                            log.warning(
                                'Buggy file system (cifs with "unix" option?) reported directory'
                                ' %s as file.', filepath)
                            _REPORTED_BUGGY_FS = True
                        # Still have to traverse this path since walk() did not.
                        size += get_dir_size(filepath)
                    else:
                        log.exception(
                            'Caught error getting file size for %s. It will not be considered '
                            'part of the total.', filepath)
    return size


class LazyMounts(Sequence[str]):
    def __init__(self) -> None:
        self._mounts: list[str] | None = None

    @staticmethod
    def _read() -> list[str]:
        return [x.split()[1] for x in Path('/proc/mounts').read_text(encoding='utf-8').splitlines()]

    def initialize(self) -> None:
        if self._mounts is None:
            self.reload()

    def reload(self) -> None:
        self._mounts = self._read()

    @property
    def mounts(self) -> list[str]:
        self.initialize()
        assert self._mounts is not None
        return self._mounts

    @override
    @overload
    def __getitem__(self, index_or_slice: int) -> str:  # pragma: no cover
        ...

    @override
    @overload
    def __getitem__(self, index_or_slice: slice) -> list[str]:  # pragma: no cover
        ...

    @override
    def __getitem__(self, index_or_slice: int | slice) -> str | list[str]:
        self.initialize()
        assert self._mounts is not None
        return self._mounts[index_or_slice]

    @override
    def __len__(self) -> int:
        self.initialize()
        assert self._mounts is not None
        return len(self._mounts)


ISO_MAX_VOLID_LENGTH = 32
MOUNTS = LazyMounts()


def is_cross_fs(dir_: str) -> bool:
    """Check if the directory is on a different file system."""
    return dir_ in MOUNTS


_DiscType = Literal['CD-R', 'DVD-R', 'DVD-R DL', 'BD-R', 'BD-R DL', 'BD-R XL (100 GB)',
                    'BD-R XL (128 GB)']


@cache
def get_disc_type(total: int) -> _DiscType:  # noqa: PLR0911
    """
    Get disc type based on total size in bytes.

    Raises
    ------
    ValueError
        If the total size exceeds the maximum supported size.
    """
    if total <= CD_R_BYTES_ADJUSTED:
        return 'CD-R'
    if total <= DVD_R_SINGLE_LAYER_SIZE_BYTES:
        return 'DVD-R'
    if total <= DVD_R_DUAL_LAYER_SIZE_BYTES_ADJUSTED:
        return 'DVD-R DL'
    if total <= BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED:
        return 'BD-R'
    if total <= BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED:
        return 'BD-R DL'
    if total <= BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED:
        return 'BD-R XL (100 GB)'
    if total <= BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED:
        return 'BD-R XL (128 GB)'
    msg = 'Disc size exceeds maximum supported size.'
    raise ValueError(msg)


@cache
def path_list_first_component(line: str) -> str:
    return re.split(r'(?<!\\)=', line, maxsplit=1)[0].replace('\\=', '=')


class WriteSpeeds(NamedTuple):
    """Write speeds for different disc types."""
    cd: int = 24
    """CD-R write speed."""
    dvd: int = 8
    """DVD-R write speed."""
    dvd_dl: float = 8
    """DVD-R DL write speed."""
    bd: int = 4
    """BD-R write speed."""
    bd_dl: int = 6
    """BD-R DL write speed."""
    bd_tl: int = 4
    """BD-R TL write speed."""
    bd_xl: int = 4
    """BD-R XL write speed."""
    def get_speed(self, disc_type: _DiscType) -> int | float:  # noqa: PLR0911
        """
        Get the write speed for the given disc type.

        Raises
        ------
        ValueError
            If the disc type is unknown.
        """
        if disc_type == 'CD-R':
            return self.cd
        if disc_type == 'DVD-R':
            return self.dvd
        if disc_type == 'DVD-R DL':
            return self.dvd_dl
        if disc_type == 'BD-R':
            return self.bd
        if disc_type == 'BD-R DL':
            return self.bd_dl
        if disc_type == 'BD-R XL (100 GB)':
            return self.bd_tl
        if disc_type == 'BD-R XL (128 GB)':
            return self.bd_xl
        msg = f'Unknown disc type: {disc_type}'  # type: ignore[unreachable]
        raise ValueError(msg)


class DirectorySplitter:
    """Split directories into sets for burning to disc."""
    def __init__(self,
                 path: os.PathLike[str] | str,
                 prefix: str,
                 delete_command: str = 'trash',
                 drive: os.PathLike[str] | str = '/dev/sr0',
                 output_dir: os.PathLike[str] | str = '.',
                 prefix_parts: tuple[str, ...] | None = None,
                 preparer: str | None = None,
                 publisher: str | None = None,
                 starting_index: int = 1,
                 write_speeds: WriteSpeeds | None = None,
                 *,
                 cross_fs: bool = False,
                 labels: bool = False) -> None:
        self._cross_fs = cross_fs
        self._current_set: list[str] = []
        self._delete_command = delete_command
        self._drive = drive or Path('/dev/sr0')
        # mogrify internally uses Inkscape for SVG to PNG conversion.
        self._has_mogrify = (False if not labels else (shutil.which('mogrify') is not None
                                                       and shutil.which('inkscape') is not None))
        self._l_path = len(str(Path(path).resolve(strict=True).parent))
        self._next_total = 0
        self._output_dir_p = Path(output_dir)
        self._path = Path(path)
        self._prefix = prefix
        self._prefix_parts = prefix_parts or (prefix,)
        self._sets: list[list[str]] = []
        self._size = 0
        self._starting_index = starting_index
        self._target_size = BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED
        self._total = 0
        self._cached_get_dir_size = cache(get_dir_size)
        self._cached_get_file_size = cache(get_file_size)
        self._write_speeds = write_speeds or WriteSpeeds()
        self._preparer = preparer
        self._publisher = publisher

    def _reset(self) -> None:
        self._target_size = BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED
        self._current_set = []
        self._total = 0

    def _too_large(self) -> None:
        self._append_set()
        self._reset()
        self._next_total = self._size

    def _append_set(self) -> None:
        if self._current_set:
            quote(f'dev={self._drive}')
            index = len(self._sets) + self._starting_index
            fn_prefix = f'{self._prefix}-{index:03d}'
            orig_vol_id = volid = f'{self._prefix}-{index:02d}'
            if len(volid) > ISO_MAX_VOLID_LENGTH:
                volid = f'{volid[:29]}-{index:02d}'
            output_dir = self._output_dir_p / fn_prefix
            output_dir.mkdir(parents=True, exist_ok=True)
            iso_file = str(output_dir / f'{fn_prefix}.iso')
            list_txt_file = f'{output_dir / orig_vol_id}.list.txt'
            pl_filename = f'{fn_prefix}.path-list.txt'
            sh_filename = f'generate-{fn_prefix}.sh'
            sha256_filename = f'{iso_file}.sha256sum'
            tree_txt_file = f'{output_dir / orig_vol_id}.tree.txt'
            metadata_filename = f'{output_dir / orig_vol_id}.metadata.json'
            log.debug('Total: %s', convert_size_bytes_to_string(self._total))
            pl_file = output_dir / pl_filename
            pl_file.write_text('\n'.join(self._current_set) + '\n', encoding='utf-8')
            label_file = output_dir / f'{fn_prefix}.png'
            disc_type = get_disc_type(self._total)
            speed = self._write_speeds.get_speed(disc_type)
            speed_s = f'{speed:.1f}' if isinstance(speed, float) else str(speed)
            special_args = []
            if self._preparer:
                special_args.append(f'-preparer {quote(self._preparer)}')
            if self._publisher:
                special_args.append(f'-publisher {quote(self._publisher)}')
            gimp_script_fu = f"""(define (print-label filename)
  (let* ((image (car (gimp-file-load RUN-INTERACTIVE filename filename))))
  (file-print-gtk #:run-mode RUN-INTERACTIVE #:image image)))
(print-label "{label_file}")""".replace('\n', '')
            delete_command = (
                f'{self._delete_command} {shlex.join(y.rsplit("=", 1)[-1] for y in self._current_set)}'  # noqa: E501
                if self._delete_command else '')
            sh_file = (output_dir / sh_filename)
            template = _jinja_env.get_template('process.sh.j2')
            sh_file.write_text(template.render(delete_command=delete_command,
                                               disc_type=disc_type,
                                               drive=quote(str(self._drive)),
                                               gimp_script_fu=quote(gimp_script_fu),
                                               iso_file=quote(iso_file),
                                               label_file=quote(str(label_file)),
                                               list_txt_file=quote(list_txt_file),
                                               metadata_filename=quote(metadata_filename),
                                               pl_file=quote(str(pl_file)),
                                               sha256_file=quote(sha256_filename),
                                               size_str=convert_size_bytes_to_string(self._total),
                                               size_bytes_formatted=f'{self._total:,}',
                                               special_args=' '.join(special_args),
                                               speed_s=quote(speed_s),
                                               tree_txt_file=quote(tree_txt_file),
                                               volid=quote(volid)) + '\n',
                               encoding='utf-8')
            sh_file.chmod(0o755)
            log.debug('%s total: %s', fn_prefix, convert_size_bytes_to_string(self._total))
            if self._has_mogrify:
                log.debug('Creating label for "%s".', orig_vol_id)
                l_common_prefix = len(commonprefix(self._current_set))
                text = f'{orig_vol_id} || ' + ' | '.join(
                    sorted(
                        path_list_first_component(x[l_common_prefix + 1:])
                        for x in self._current_set if x.strip()))
                write_spiral_text_png(label_file, text)
            self._sets.append(self._current_set)

    def split(self) -> None:
        """Split the directory into sets."""
        cmd = ('find', str(Path(self._path).resolve(strict=True)), '-maxdepth', '1', '(', '-name',
               '.Trash-*', '-o', '-name', 'Trash', '-o', '-name', '.Trash', '-o', '-name',
               '.directory', ')', '-prune', '-o', '-print')
        log.debug('Running %s', ' '.join(quote(x) for x in cmd))
        for dir_ in sorted(sorted(
                sp.run(cmd, check=True, text=True, capture_output=True).stdout.splitlines()[1:]),
                           key=lambda x: not isdir(x)):  # noqa: PTH112
            if not self._cross_fs and is_cross_fs(dir_):
                log.debug('Not processing %s because it is another file system.', dir_)
                continue
            log.debug('Calculating size: %s', dir_)
            type_ = 'Directory'
            try:
                self._size = self._cached_get_dir_size(dir_)
            except NotADirectoryError:
                type_ = 'File'
                try:
                    self._size = self._cached_get_file_size(dir_)
                except OSError:
                    continue
            self._next_total = self._total + self._size
            log.debug('%s: %s - %s', type_, dir_, convert_size_bytes_to_string(self._size))
            log.debug('Current total: %s / %s', convert_size_bytes_to_string(self._next_total),
                      convert_size_bytes_to_string(self._target_size))
            if self._next_total > self._target_size:
                log.debug('Current set with %s exceeds target size.', dir_)
                if self._target_size == BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED:
                    log.debug('Trying quad layer.')
                    self._target_size = BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED
                    if self._next_total > self._target_size:
                        log.debug('Still too large. Appending to next set.')
                        self._too_large()
                else:
                    self._too_large()
            if (self._next_total > self._target_size
                    and self._target_size == BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED
                    and self._next_total > BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED):
                if type_ == 'File':
                    log.warning(
                        'File %s too large for largest Blu-ray disc. It will not be processed.',
                        dir_)
                    continue
                log.debug('Directory %s too large for Blu-ray. Splitting separately.', dir_)
                suffix = Path(dir_).name
                DirectorySplitter(dir_,
                                  f'{self._prefix}-{suffix}',
                                  cross_fs=self._cross_fs,
                                  delete_command=self._delete_command,
                                  drive=self._drive,
                                  labels=self._has_mogrify,
                                  output_dir=self._output_dir_p,
                                  prefix_parts=(*self._prefix_parts, suffix),
                                  preparer=self._preparer,
                                  publisher=self._publisher,
                                  starting_index=self._starting_index,
                                  write_speeds=self._write_speeds).split()
                self._reset()
                continue
            self._total = self._next_total
            fixed = dir_[self._l_path + 1:].replace('=', '\\=')
            self._current_set.append(f'{fixed}={dir_}')
        self._append_set()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from gendisc.main import genlabel_main, main

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_main_success(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('gendisc.main.Path.mkdir')
    mocker.patch('gendisc.main.keep.running')
    mocker.patch('gendisc.main.DirectorySplitter')
    result = runner.invoke(main, ('test_path', '-D', '/dev/sr0', '-o', 'output_dir', '-i', '1'))
    assert result.exit_code == 0


def test_main_debug_logging(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('gendisc.main.Path.mkdir')
    mocker.patch('gendisc.main.keep.running')
    mocker.patch('gendisc.main.DirectorySplitter')
    mock_logging = mocker.patch('gendisc.main.setup_logging')
    runner.invoke(main, ['test_path', '--debug'])
    mock_logging.assert_called_once_with(debug=True)


def test_main_default_values(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('gendisc.main.Path.mkdir')
    mocker.patch('gendisc.main.keep.running')
    mock_splitter = mocker.patch('gendisc.main.DirectorySplitter')
    runner.invoke(main, ['test_path'])
    mock_splitter.assert_called_once()
    args, kwargs = mock_splitter.call_args
    assert args[0].name == 'test_path'
    assert args[1] == 'test_path'
    assert kwargs['delete_command'] == 'trash'
    assert kwargs['drive'] == Path('/dev/sr0')
    assert kwargs['output_dir'] == Path.cwd()
    assert kwargs['cross_fs'] is False
    assert kwargs['labels'] is True


def test_main_delete_option(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('gendisc.main.Path.mkdir')
    mocker.patch('gendisc.main.keep.running')
    mock_splitter = mocker.patch('gendisc.main.DirectorySplitter')
    runner.invoke(main, ['test_path', '--delete'])
    mock_splitter.assert_called_once()
    _args, kwargs = mock_splitter.call_args
    assert kwargs['delete_command'] == 'rm -rf'


def test_genlabel_main_svg_output(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_svg = mocker.patch('gendisc.main.write_spiral_text_svg')
    result = runner.invoke(genlabel_main,
                           ['Hello', 'World', '--svg', '-o', 'label.svg', '-w', '500', '-f', '20'])
    assert result.exit_code == 0
    mock_svg.assert_called_once()
    args = mock_svg.call_args[0]
    assert args[0].name.endswith('.svg')
    assert 'Hello World' in args


def test_genlabel_main_png_output(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_png = mocker.patch('gendisc.main.write_spiral_text_png')
    result = runner.invoke(genlabel_main, ['Test', '-o', 'label.png', '-w', '400', '--dpi', '300'])
    assert result.exit_code == 0
    mock_png.assert_called_once()
    args = mock_png.call_args[0]
    assert args[0].name.endswith('.png')
    assert 'Test' in args


def test_genlabel_main_with_center_and_view_box(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_png = mocker.patch('gendisc.main.write_spiral_text_png')
    result = runner.invoke(genlabel_main, [
        'Centred', '-o', 'centred.png', '-c', '100', '100', '-V', '0', '0', '400', '400', '-w',
        '400'
    ])
    assert result.exit_code == 0
    mock_png.assert_called_once()
    args = mock_png.call_args[0]
    # Check that Point is passed for center
    assert hasattr(args[7], 'x')
    assert hasattr(args[7], 'y')
    # Check that view_box is passed
    assert args[4] == (0, 0, 400, 400)


def test_genlabel_main_keep_svg_flag(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_png = mocker.patch('gendisc.main.write_spiral_text_png')
    runner.invoke(genlabel_main, ['KeepSVG', '-o', 'keep.png', '--keep-svg'])
    mock_png.assert_called_once()
    kwargs = mock_png.call_args.kwargs
    assert kwargs.get('keep') is True


def test_genlabel_main_font_size_and_theta(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_png = mocker.patch('gendisc.main.write_spiral_text_png')
    runner.invoke(genlabel_main, ['FontTest', '-o', 'font.png', '-f', '22', '-t', '45'])
    mock_png.assert_called_once()
    args = mock_png.call_args[0]
    assert args[6] == 22  # font_size
    assert args[12] == 45  # theta_step



"""Constants."""
from __future__ import annotations

__all__ = ('BLURAY_DUAL_LAYER_SIZE_BYTES', 'BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_QUADRUPLE_LAYER_SIZE_BYTES', 'BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_SINGLE_LAYER_SIZE_BYTES', 'BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_TRIPLE_LAYER_SIZE_BYTES', 'BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED',
           'CD_R_BYTES', 'CD_R_BYTES_ADJUSTED', 'DVD_R_DUAL_LAYER_SIZE_BYTES',
           'DVD_R_SINGLE_LAYER_SIZE_BYTES')

CD_R_BYTES = 737_280_000
"""700 MB CD-R size in bytes."""
CD_R_BYTES_ADJUSTED = CD_R_BYTES - 5_242_880
"""700 MB CD-R size in bytes with 5 MiB free space."""
DVD_R_SINGLE_LAYER_SIZE_BYTES = 4_707_319_808
"""DVD-R size in bytes."""
DVD_R_DUAL_LAYER_SIZE_BYTES = 8_543_666_176
"""DVD-R DL size in bytes."""
FREE_SPACE_BYTES = 52_428_800
"""50 MiB free space."""
DVD_R_SINGLE_LAYER_SIZE_BYTES_ADJUSTED = DVD_R_SINGLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""DVD-R size in bytes with 50 MiB left empty."""
DVD_R_DUAL_LAYER_SIZE_BYTES_ADJUSTED = DVD_R_DUAL_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""DVD-R DL size in bytes with 50 MiB left empty."""
BLURAY_SINGLE_LAYER_SIZE_BYTES = 25_025_314_816
"""25 GB Blu-Ray disc size."""
BLURAY_DUAL_LAYER_SIZE_BYTES = 50_050_629_632
"""50 GB Blu-Ray disc size."""
BLURAY_TRIPLE_LAYER_SIZE_BYTES = 100_103_356_416
"""100 GB Blu-Ray disc size."""
BLURAY_QUADRUPLE_LAYER_SIZE_BYTES = 128_001_769_472
"""128 GB Blu-Ray disc size in bytes."""
BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_SINGLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""25 GB disc size in bytes with 50 MiB left empty."""
BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_DUAL_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""50 GB disc size in bytes with 50 MiB left empty."""
BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_TRIPLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""100 GB disc size in bytes with 50 MiB left empty."""
BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_QUADRUPLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""128 GB disc size in bytes with 50 MiB left empty."""



"""Disc label generation."""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from shlex import quote
from typing import (
    TYPE_CHECKING,
    SupportsFloat,
    SupportsIndex,
    TypeAlias,
)
from xml.sax.saxutils import escape
import logging
import math
import shutil
import subprocess as sp

if TYPE_CHECKING:
    import os

__all__ = ('MogrifyNotFound', 'Point', 'create_spiral_path', 'create_spiral_text_svg',
           'write_spiral_text_png', 'write_spiral_text_svg')

log = logging.getLogger(__name__)


@dataclass
class Point:
    """Point class for SVG paths."""
    x: float
    """X coordinate."""
    y: float
    """Y coordinate."""


def _line_intersection(m1: float, b1: float, m2: float, b2: float) -> Point:
    """
    Find the intersection of two lines.

    Parameters
    ----------
    m1 : int
        Slope of the first line.
    b1 : int
        Y-intercept of the first line.
    m2 : int
        Slope of the second line.
    b2 : int
        Y-intercept of the second line.

    Returns
    -------
    Point
        X and Y coordinates of the intersection point.

    Raises
    ------
    ValueError
        If the lines are parallel and do not intersect.
    """
    if m1 == m2:
        msg = 'Lines are parallel and do not intersect.'
        raise ValueError(msg)
    x = (b2 - b1) / (m1 - m2)
    y = m1 * x + b1
    return Point(x, y)


def _p_str(point: Point) -> str:
    """
    Convert a point to a string.

    Parameters
    ----------
    point : tuple[int, int]
        The point to convert.

    Returns
    -------
    str
        The point as a string.
    """
    return f'{point.x},{point.y} '


_SupportsFloatOrIndex: TypeAlias = SupportsFloat | SupportsIndex
DEFAULT_START_RADIUS = 0
DEFAULT_SPACE_PER_LOOP = 40
DEFAULT_START_THETA = -6840
DEFAULT_END_THETA = 0
DEFAULT_THETA_STEP = 30
DEFAULT_FONT_SIZE = 32
DEFAULT_WIDTH_HEIGHT = 800
DEFAULT_DPI = 300


def create_spiral_path(center: Point | None = None,
                       start_radius: float = DEFAULT_START_RADIUS,
                       space_per_loop: float = DEFAULT_SPACE_PER_LOOP,
                       start_theta: _SupportsFloatOrIndex = DEFAULT_START_THETA,
                       end_theta: _SupportsFloatOrIndex = DEFAULT_END_THETA,
                       theta_step: _SupportsFloatOrIndex = DEFAULT_THETA_STEP) -> str:
    """
    Get a path string for a spiral in a SVG file.

    Defaults to creating a spiral that starts at the outside and goes inwards.

    Algorithm borrowed from `How to make a spiral in SVG? <https://stackoverflow.com/a/49099258/374110>`_.

    Parameters
    ----------
    center : Point
        The center of the spiral.
    start_radius : float
        The starting radius of the spiral.
    space_per_loop : float
        The space between each loop of the spiral.
    start_theta : float
        The starting angle of the spiral in degrees. Use a negative number to have the spiral start
        at the outside and go inwards.
    end_theta : float
        The ending angle of the spiral in degrees.
    theta_step : float
        The step size of the angle in degrees.

    Returns
    -------
    str
        The path string for the spiral. Goes inside a ``<path>`` in the ``d`` attribute.
    """
    log.debug(
        'Creating spiral path with center %s, start radius %.2f, space per loop %.2f, '
        'start theta %.2f, end theta %.2f, and theta step %.2f.', center, start_radius,
        space_per_loop, start_theta, end_theta, theta_step)
    center = center or Point(DEFAULT_WIDTH_HEIGHT, DEFAULT_WIDTH_HEIGHT)
    # Rename spiral parameters for the formula r = a + bθ.
    a = start_radius  # Start distance from center
    b = space_per_loop / math.pi / 2  # Space between each loop
    # Convert angles to radians.
    old_theta = new_theta = math.radians(start_theta)
    end_theta = math.radians(end_theta)
    theta_step = math.radians(theta_step)
    # Radii
    new_r = a + b * new_theta
    # Start and end points
    old_point = Point(0, 0)
    new_point = Point(center.x + new_r * math.cos(new_theta),
                      center.y + new_r * math.sin(new_theta))
    # Slopes of tangents
    new_slope = ((b * math.sin(old_theta) + (a + b * new_theta) * math.cos(old_theta)) /
                 (b * math.cos(old_theta) - (a + b * new_theta) * math.sin(old_theta)))
    paths = f'M {_p_str(new_point)} Q '
    while old_theta < end_theta - theta_step:
        old_theta = new_theta
        new_theta += theta_step
        old_r = new_r
        new_r = a + b * new_theta
        old_point.x = new_point.x
        old_point.y = new_point.y
        new_point.x = center.x + new_r * math.cos(new_theta)
        new_point.y = center.y + new_r * math.sin(new_theta)
        # Slope calculation with the formula
        # m := (b * sin(θ) + (a + b * θ) * cos(θ)) / (b * cos(θ) - (a + b * θ) * sin(θ))
        a_plus_b_theta = a + b * new_theta
        old_slope = new_slope
        new_slope = ((b * math.sin(new_theta) + a_plus_b_theta * math.cos(new_theta)) /
                     (b * math.cos(new_theta) - a_plus_b_theta * math.sin(new_theta)))
        old_intercept = -(old_slope * old_r * math.cos(old_theta) - old_r * math.sin(old_theta))
        new_intercept = -(new_slope * new_r * math.cos(new_theta) - new_r * math.sin(new_theta))
        control_point = _line_intersection(old_slope, old_intercept, new_slope, new_intercept)
        # Offset the control point by the center offset.
        control_point.x += center.x
        control_point.y += center.y
        paths += f'{_p_str(control_point)}{_p_str(new_point)} '
    return paths.strip()


def create_spiral_text_svg(text: str,
                           width: int = DEFAULT_WIDTH_HEIGHT,
                           height: int | None = None,
                           view_box: tuple[int, int, int, int] | None = None,
                           font_size: int = DEFAULT_FONT_SIZE,
                           center: Point | None = None,
                           start_radius: float = DEFAULT_START_RADIUS,
                           space_per_loop: float = DEFAULT_SPACE_PER_LOOP,
                           start_theta: _SupportsFloatOrIndex = DEFAULT_START_THETA,
                           end_theta: _SupportsFloatOrIndex = DEFAULT_END_THETA,
                           theta_step: _SupportsFloatOrIndex = DEFAULT_THETA_STEP) -> str:
    """
    Create a spiral SVG text.

    Defaults to creating a spiral that starts at the outside and goes inwards.

    Parameters
    ----------
    text: str
        The text to put in the spiral.
    width : int
        The width of the SVG.
    height : int
        The height of the SVG.
    view_box : str
        The view box of the SVG.
    font_size : int
        The font size of the text in the SVG in pixels.
    center : Point
        The center of the spiral. If not specified, it will be set to (width, width).
    start_radius : float
        The starting radius of the spiral.
    space_per_loop : float
        The space between each loop of the spiral.
    start_theta : float
        The starting angle of the spiral in degrees. Use a negative number to have the spiral start
        at the outside and go inwards.
    end_theta : float
        The ending angle of the spiral in degrees.
    theta_step : float
        The step size of the angle in degrees.

    Returns
    -------
    str
        The SVG string for the spiral.
    """
    text = escape(text)
    center = center or Point(width, width)
    height = height or width
    view_box_s = ' '.join((str(x) for x in view_box) if view_box else ('0', '0', str(width * 2),
                                                                       str(height * 2)))
    path_d = create_spiral_path(center, start_radius, space_per_loop, start_theta, end_theta,
                                theta_step)
    return f"""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg xmlns="http://www.w3.org/2000/svg" width="{width}" height="{height}" viewBox="{view_box_s}">
  <style>
    .small {{
      font: {font_size}px sans-serif;
    }}
  </style>
  <path id="spiral" d="{path_d}" fill="none" stroke="black" stroke-width="0" />
  <text>
    <textPath href="#spiral" class="small">
    {text}
    </textPath>
  </text>
</svg>""".strip()


def write_spiral_text_svg(filename: str | os.PathLike[str],
                          text: str,
                          width: int = DEFAULT_WIDTH_HEIGHT,
                          height: int | None = None,
                          view_box: tuple[int, int, int, int] | None = None,
                          font_size: int = DEFAULT_FONT_SIZE,
                          center: Point | None = None,
                          start_radius: float = DEFAULT_START_RADIUS,
                          space_per_loop: float = DEFAULT_SPACE_PER_LOOP,
                          start_theta: _SupportsFloatOrIndex = DEFAULT_START_THETA,
                          end_theta: _SupportsFloatOrIndex = DEFAULT_END_THETA,
                          theta_step: _SupportsFloatOrIndex = DEFAULT_THETA_STEP) -> None:
    """
    Write a spiral text SVG string to a file.

    Defaults to creating a spiral that starts at the outside and goes inwards.

    Parameters
    ----------
    filename : str | os.PathLike[str]
        The filename to write the SVG to.
    text: str
        The text to put in the spiral.
    width : int
        The width of the SVG.
    height : int
        The height of the SVG.
    view_box : tuple[int, int, int, int] | None
        The view box of the SVG. If not specified, it will be set to
        ``(0, 0, width * 2, height * 2)``.
    font_size : int
        The font size of the text in the SVG in pixels.
    center : Point
        The center of the spiral.
    start_radius : float
        The starting radius of the spiral.
    space_per_loop : float
        The space between each loop of the spiral.
    start_theta : float
        The starting angle of the spiral in degrees.
    end_theta : float
        The ending angle of the spiral in degrees.
    theta_step : float
        The step size of the angle in degrees.
    """
    filename = Path(filename)
    spiral_svg = create_spiral_text_svg(text, width, height or width, view_box, font_size, center,
                                        start_radius, space_per_loop, start_theta, end_theta,
                                        theta_step)
    filename.write_text(f'{spiral_svg}\n', encoding='utf-8')


class MogrifyNotFound(FileNotFoundError):
    """Raised when ``mogrify`` is not found in ``PATH``."""
    def __init__(self) -> None:
        msg = '`mogrify` not found in PATH. Please install ImageMagick.'
        super().__init__(msg)


def write_spiral_text_png(filename: str | os.PathLike[str],
                          text: str,
                          width: int = DEFAULT_WIDTH_HEIGHT,
                          height: int | None = None,
                          view_box: tuple[int, int, int, int] | None = None,
                          dpi: int = DEFAULT_DPI,
                          font_size: int = DEFAULT_FONT_SIZE,
                          center: Point | None = None,
                          start_radius: float = DEFAULT_START_RADIUS,
                          space_per_loop: float = DEFAULT_SPACE_PER_LOOP,
                          start_theta: _SupportsFloatOrIndex = DEFAULT_START_THETA,
                          end_theta: _SupportsFloatOrIndex = DEFAULT_END_THETA,
                          theta_step: _SupportsFloatOrIndex = DEFAULT_THETA_STEP,
                          *,
                          debug: bool = False,
                          keep: bool = False) -> None:
    """
    Write a spiral text SVG string to a file.

    Defaults to creating a spiral that starts at the outside and goes inwards.

    Requires ``mogrify`` from ImageMagick to be installed and in ``PATH``.

    Parameters
    ----------
    filename : str | os.PathLike[str]
        The filename to write the PNG to.
    text: str
        The text to put in the spiral.
    width : int
        The width of the SVG.
    height : int
        The height of the SVG.
    view_box : tuple[int, int, int, int] | None
        The view box of the SVG. If not specified, it will be set to
        ``(0, 0, width * 2, height * 2)``.
    font_size : int
        The font size of the text in the SVG in pixels.
    center : Point
        The center of the spiral.
    start_radius : float
        The starting radius of the spiral.
    space_per_loop : float
        The space between each loop of the spiral.
    start_theta : float
        The starting angle of the spiral in degrees.
    end_theta : float
        The ending angle of the spiral in degrees.
    theta_step : float
        The step size of the angle in degrees.
    keep : bool
        If ``True``, keep the SVG file after conversion.

    Raises
    ------
    MogrifyNotFound
        If ``mogrify`` is not found in ``PATH``.
    FileNotFoundError
        If the PNG file could not be created.
    """
    if not (mogrify := shutil.which('mogrify')):
        raise MogrifyNotFound
    log.debug('Writing spiral text image to %s. Be patient!', filename)
    filename = Path(filename)
    svg_file = filename.with_suffix('.svg')
    write_spiral_text_svg(svg_file, text, width, height, view_box, font_size, center, start_radius,
                          space_per_loop, start_theta, end_theta, theta_step)
    size_args = ('-density', str(dpi), '-gravity', 'center', '-resize', '2800x2800', '-extent',
                 '2835x2835')
    cmd: tuple[str, ...] = (mogrify, '-comment', 'gendisc', '-colorspace', 'sRGB', '-units',
                            'PixelsPerInch', *size_args, '-background', 'none', '-format', 'png',
                            str(svg_file))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    if not filename.exists():
        msg = f'Failed to create {filename}.'
        raise FileNotFoundError(msg)
    if not keep:
        svg_file.unlink()



"""Main command."""
from __future__ import annotations

from functools import partialmethod
from pathlib import Path
import logging

from tqdm import tqdm
from wakepy import keep
import click

from .genlabel import (
    DEFAULT_DPI,
    DEFAULT_END_THETA,
    DEFAULT_FONT_SIZE,
    DEFAULT_SPACE_PER_LOOP,
    DEFAULT_START_RADIUS,
    DEFAULT_START_THETA,
    DEFAULT_THETA_STEP,
    DEFAULT_WIDTH_HEIGHT,
    Point,
    write_spiral_text_png,
    write_spiral_text_svg,
)
from .utils import DirectorySplitter, WriteSpeeds, setup_logging

__all__ = ('main',)

log = logging.getLogger(__name__)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.argument('path', type=click.Path(file_okay=False, resolve_path=True, path_type=Path))
@click.option('--cross-fs', help='Allow crossing file systems.', is_flag=True)
@click.option('-D',
              '--drive',
              default='/dev/sr0',
              help='Drive path.',
              type=click.Path(dir_okay=False, resolve_path=True, path_type=Path))
@click.option('-d', '--debug', help='Enable debug logging.', is_flag=True)
@click.option('-i',
              '--starting-index',
              default=1,
              help='Index to start with (defaults to 1).',
              metavar='INDEX',
              type=click.IntRange(1))
@click.option('-o',
              '--output-dir',
              default='.',
              help='Output directory. Will be created if it does not exist.',
              type=click.Path(file_okay=False, resolve_path=True, path_type=Path))
@click.option('-p', '--prefix', help='Prefix for volume ID and files.')
@click.option('-r', '--delete', help='Unlink instead of sending to trash.', is_flag=True)
@click.option('--no-labels', help='Do not create labels.', is_flag=True)
@click.option('--cd-write-speed', help='CD-R write speed.', type=int, default=24)
@click.option('--dvd-write-speed', help='DVD-R write speed.', type=int, default=8)
@click.option('--dvd-dl-write-speed', help='DVD-R DL write speed.', type=float, default=8)
@click.option('--bd-write-speed', help='BD-R write speed.', type=int, default=4)
@click.option('--bd-dl-write-speed', help='BD-R DL write speed.', type=int, default=6)
@click.option('--bd-tl-write-speed', help='BD-R TL write speed.', type=int, default=4)
@click.option('--bd-xl-write-speed', help='BD-R XL write speed.', type=int, default=4)
@click.option('--preparer', help='Preparer string (128 characters).', type=str)
@click.option('--publisher', help='Publisher string (128 characters).', type=str)
def main(path: Path,
         output_dir: Path,
         drive: Path,
         preparer: str | None = None,
         publisher: str | None = None,
         prefix: str | None = None,
         starting_index: int = 0,
         cd_write_speed: int = 24,
         dvd_write_speed: int = 8,
         dvd_dl_write_speed: float = 8,
         bd_write_speed: int = 4,
         bd_dl_write_speed: int = 6,
         bd_tl_write_speed: int = 4,
         bd_xl_write_speed: int = 4,
         *,
         cross_fs: bool = False,
         debug: bool = False,
         delete: bool = False,
         no_labels: bool = False) -> None:
    """Make a file listing filling up discs."""
    setup_logging(debug=debug)
    if debug:
        tqdm.__init__ = partialmethod(  # type: ignore[assignment,method-assign]
            tqdm.__init__, disable=True)
    output_dir_p = Path(output_dir).resolve()
    output_dir_p.mkdir(parents=True, exist_ok=True)
    with keep.running():
        DirectorySplitter(path,
                          prefix or path.name,
                          cross_fs=cross_fs,
                          delete_command='rm -rf' if delete else 'trash',
                          drive=drive,
                          labels=not no_labels,
                          output_dir=output_dir_p,
                          preparer=preparer,
                          publisher=publisher,
                          starting_index=starting_index,
                          write_speeds=WriteSpeeds(cd=cd_write_speed,
                                                   dvd=dvd_write_speed,
                                                   dvd_dl=dvd_dl_write_speed,
                                                   bd=bd_write_speed,
                                                   bd_dl=bd_dl_write_speed,
                                                   bd_tl=bd_tl_write_speed,
                                                   bd_xl=bd_xl_write_speed)).split()


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.argument('text', nargs=-1)
@click.option('-E', '--end-theta', help='End theta.', type=float, default=0)
@click.option('-H', '--height', help='Height of the image.', type=int)
@click.option('-S',
              '--space-per-loop',
              help='Space per loop.',
              type=float,
              default=DEFAULT_SPACE_PER_LOOP)
@click.option('-T', '--start-theta', help='Start theta.', type=float, default=DEFAULT_START_THETA)
@click.option('-V',
              '--view-box',
              help='SVG view box.',
              type=click.Tuple((int, int, int, int)),
              required=False)
@click.option('--dpi', help='Dots per inch.', type=int, default=DEFAULT_DPI)
@click.option('--keep-svg', help='When generating the PNG, keep the SVG file.', is_flag=True)
@click.option('-c', '--center', help='Center of the spiral.', type=click.Tuple((float, float)))
@click.option('-d', '--debug', help='Enable debug logging.', is_flag=True)
@click.option('-f', '--font-size', help='Font size.', type=float, default=DEFAULT_FONT_SIZE)
@click.option('-g', '--svg', help='Output SVG.', is_flag=True)
@click.option('-o',
              '--output',
              help='Output file name.',
              type=click.Path(path_type=Path, dir_okay=False),
              default='out.png')
@click.option('-r',
              '--start-radius',
              help='Start radius.',
              type=float,
              default=DEFAULT_START_RADIUS)
@click.option('-t', '--theta-step', help='Theta step.', type=float, default=DEFAULT_THETA_STEP)
@click.option('-w',
              '--width',
              help='Width of the image.',
              type=click.IntRange(1, 10000),
              default=DEFAULT_WIDTH_HEIGHT)
def genlabel_main(text: tuple[str, ...],
                  output: Path,
                  center: tuple[float, float] | None = None,
                  dpi: int = DEFAULT_DPI,
                  end_theta: float = DEFAULT_END_THETA,
                  font_size: int = DEFAULT_FONT_SIZE,
                  height: int | None = None,
                  space_per_loop: float = DEFAULT_SPACE_PER_LOOP,
                  start_radius: int = DEFAULT_START_RADIUS,
                  start_theta: float = DEFAULT_START_THETA,
                  theta_step: float = DEFAULT_THETA_STEP,
                  view_box: tuple[int, int, int, int] | None = None,
                  width: int = DEFAULT_WIDTH_HEIGHT,
                  *,
                  debug: bool = False,
                  keep_svg: bool = False,
                  svg: bool = False) -> None:
    """Generate an image intended for printing on disc consisting of text in a spiral."""
    setup_logging(debug=debug)
    if svg:
        write_spiral_text_svg(output.with_suffix('.svg'), ' '.join(text), width, height, view_box,
                              font_size,
                              Point(*center) if center else None, start_radius, space_per_loop,
                              start_theta, end_theta, theta_step)
    else:
        write_spiral_text_png(output,
                              ' '.join(text),
                              width,
                              height,
                              view_box,
                              dpi,
                              font_size,
                              Point(*center) if center else None,
                              start_radius,
                              space_per_loop,
                              start_theta,
                              end_theta,
                              theta_step,
                              keep=keep_svg,
                              debug=debug)
